#include <string.h>
#include <iostream>
#include <vector>
#include <GLUT/glut.h>
#include "Image.h"
#include "ImageRenderer.h"
#include "AFMM.h"
#include "Flags.h"


using namespace std;

Image<float>*		  input_image		 =0;				//Input image (grayscale)
Image<float>*		  thresholded_image  =0;				//Thresholding of input image
Image<float>*		  distance_transform =0;				//Distance-transform of boundary of thresholded image
Image<float>*		  feature_transform	 =0;				//Feature transform of boundary of thresholded image
Image<float>*		  importance		 =0;				//Skeleton importance
Image<float>*		  skeleton			 =0;				//Simplified skeleton
ImageRenderer<float>* renderer			 =0;				//Image renderer

float				  k = -100;								//Grayscale threshold (see FLAGS class)
float				  threshold = 4;						//Skeleton-importance threshold
	
vector<Image<float>*> images;								//Set of all images (with descriptive names) to display
vector<const char*>   image_names;
int					  f_display = 0;						//Current image to display




inline float wrapDistance(float cd,float length)			//Compute distances on boundary in wrap-around fashion
{ return min(fabs(float(cd)),fabs(length-fabs(float(cd)))); }



void computeSkeletonImportance(FLAGS* f,Image<float>* FT,Image<float>* importance,float boundaryLength)
{															//Compute skeleton importance from feature transform

   for(int j=0;j<importance->getHeight();j++)				// 
     for(int i=0;i<importance->getWidth();i++)				//	
     {
		 float& imp = (*importance)(j,i);
		 if (f->faraway(i,j))								//Only compute importance for pixels inside the shape
		 {													//(for which we computed the feature transform)
			float ux = (*FT)(j,i+1) - (*FT)(j,i);
			float uy = (*FT)(j+1,i) - (*FT)(j,i);
			float d1 = wrapDistance(ux,boundaryLength);
			float d2 = wrapDistance(uy,boundaryLength);
			imp = max(d1,d2);
		 }
		 else imp = 0;										//No skeleton-pixels exist outside the shape	
     }
}



void computeSimplifiedSkeleton(Image<float>* importance,Image<float>* skeleton,float threshold)
{															//Compute simplified skeleton by thresholding importance
   for(int j=0;j<importance->getHeight();j++)				//
     for(int i=0;i<importance->getWidth();i++)				//	
     {
		 (*skeleton)(j,i) = ((*importance)(j,i)>threshold)? 1:0;
	 }
}	 	 


void computeDT(Image<float>& input_image,Image<float>& thresholded_image,Image<float>& distance_transform,float k)
{															//Threshold the input image with value k, to generate a binary image.
															//Next, compute the distance-transform of the boundary of this 
   thresholded_image = input_image;							//binary image, using the Fast Marching Method.

   FLAGS flags(thresholded_image,k); 
   
   FLAGS flags_input = flags;								//Make a copy of the classified image. We need this to determine
															//which pixels are inside the shape AFTER the AFMM has executed,
															//since AFMM modifies 'flags'.
   distance_transform = thresholded_image;
      
   AugmentedFastMarchingMethod fmm(&distance_transform,&flags,feature_transform);		
   fmm.setMethod(AugmentedFastMarchingMethod::AFMM_STAR);	//Compute distance and feature transform
   int iterations = fmm.execute();							//Compute the distance transform
   cout<<"FMM iterations: "<<iterations<<endl;				//Display how many iterations we needed

   float length = fmm.getLength();
   
   computeSkeletonImportance(&flags_input,feature_transform,importance,length);
															//Compute the skeleton importance
   computeSimplifiedSkeleton(importance,skeleton,threshold);//Compute the thresholded skeleton	
}





void viewing(int width, int height)							//GLUT window-reshape callback
{
	glutReshapeWindow(width,height);						//Initialze the OpenGL modelview and projection
	glMatrixMode(GL_MODELVIEW);								//matrices so as to have a 1-to-1 correspondence
	glLoadIdentity();										//between world coordinates and pixels
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluOrtho2D(0,width,0,height);
	glViewport(0,0,width,height);
}



void keyboard(unsigned char ch,int,int)						//GLUt keyboard callback
{
    switch(ch)
    {
      case ' ':												//Cycle through the available images
	  { 
	    f_display = (++f_display)%(images.size()); 
		cout<<"Showing: "<<image_names[f_display]<<endl;
		break; 
	  }
	  case '+':												//Increase threshold to generate new images
	  case '=':
	  {
		k++;		
	    computeDT(*input_image,*thresholded_image,*distance_transform,k);
		break;
	  }	
	  case '-':												//Decrease threshold to generate new images
	  case '_':
	  {
		k--;		
	    computeDT(*input_image,*thresholded_image,*distance_transform,k);
		break;
	  }	
      case 'c':												//Display images either as grayscale or color-mapped
	  case 'C':
	    renderer->setDrawColored(!renderer->getDrawColored()); 
		break;
	  case '<':
	  case ',':
	    threshold--; if (threshold<2) threshold=2;	
		computeSimplifiedSkeleton(importance,skeleton,threshold);
		break;
	  case '>':
	  case '.':
	    threshold++; 	
		computeSimplifiedSkeleton(importance,skeleton,threshold);
		break;		
    }

    glutPostRedisplay();
}


void display()												//GLUT display callback
{
  renderer->draw(*images[f_display]);						//Display currently selected image
}




int main(int argc,char* argv[])
{
   cout<<"3. Multiscale skeletons"<<endl;
   cout<<"In the graphics window, press"<<endl;
   cout<<"   space: cycle through the images to display"<<endl;
   cout<<"   +,-:   increase/decrease image threshold"<<endl;
   cout<<"   <,>:   increase/decrease skeleton threshold"<<endl;
   cout<<"          (visible in the simplified skeleton image)"<<endl;
   cout<<"   c,C:   show images as grayscale or color-mapped"<<endl;
   
   
   glutInit(&argc,argv);
   const char* filename = (argc>1)? argv[1] : "DATA/skinlesion.ppm";
   
   if (strstr(filename,".pgm"))	
      input_image = Image<float>::readPGM(filename);				//1. Read an input image
   else if (strstr(filename,".ppm"))								//   If color, convert to luminance
      input_image = Image<float>::readPPM(filename);
	  	  
   if (!input_image) 
   { cout<<"Can not open file: "<<filename<<endl; return 1; }

   thresholded_image  = new Image<float>(*input_image);				//2. Allocate all our images
   distance_transform = new Image<float>(*input_image);				// 
   feature_transform  = new Image<float>(*input_image);
   importance         = new Image<float>(*input_image);
   skeleton           = new Image<float>(*input_image);

   computeDT(*input_image,*thresholded_image,*distance_transform,k);//3. Compute the distance transform,
																	//   skeleton-importance, and simplified skeleton
																	//   for the thresholded shape

   glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE);
   glutInitWindowSize(input_image->getWidth(),input_image->getHeight());
   glutCreateWindow("3. Multiscale skeletons");
   
																	//4. Gather all images to visualize
   images.push_back(input_image);		 image_names.push_back("Input image");
   images.push_back(thresholded_image);  image_names.push_back("Thresholded image");
   images.push_back(distance_transform); image_names.push_back("Distance transform");
   images.push_back(feature_transform);	 image_names.push_back("Feature transform");
   images.push_back(importance);		 image_names.push_back("Skeleton importance");
   images.push_back(skeleton);			 image_names.push_back("Simplified skeleton");
   
   renderer = new ImageRenderer<float>();							//5. Create an image renderer
   renderer->setDrawColored(true);									//   Use color-mapping, since most images we have
																	//   are not normalized in [0..255], like e.g. PGM/PPM ones

   glutKeyboardFunc(keyboard);
   glutDisplayFunc(display);
   glutReshapeFunc(viewing);
   glutMainLoop();
   
   return 0;
}


